"""
BornAgain functional test.
Test of rotation/position of particle composition.
The composition consists of two boxes made of the same material.
It is compared against reference single box made of the same material.
Composition might be rotated to get reference shape.
Both, reference box and composition are placed in the center of middle layer of 3 layers system.
"""

from __future__ import print_function
import os, sys, unittest

import utils
from bornagain import *

layer_thickness = 100.0
com_length = 50.0
com_width = 20.0
com_height = 10.0
particle_material = HomogeneousMaterial("Ag", 1.245e-5, 5.419e-7)


class TransformBoxCompositionTest(unittest.TestCase):
    def get_sample(self, particle):
        mAmbience = HomogeneousMaterial("Vacuum", 0, 0)
        mMiddle = HomogeneousMaterial("Teflon", 2.900e-6, 6.019e-9)
        mSubstrate = HomogeneousMaterial("Substrate", 3.212e-6, 3.244e-8)

        layout = ParticleLayout()
        layout.addParticle(particle)

        vacuum_layer = Layer(mAmbience)
        middle_layer = Layer(mMiddle, layer_thickness)
        # vacuum_layer.addLayout(layout)
        middle_layer.addLayout(layout)
        substrate = Layer(mSubstrate)

        multi_layer = MultiLayer()
        multi_layer.addLayer(vacuum_layer)
        multi_layer.addLayer(middle_layer)
        multi_layer.addLayer(substrate)
        return multi_layer

    def get_result(self, particle):
        sample = self.get_sample(particle)
        simulation = utils.get_simulation_MiniGISAS(sample)
        simulation.runSimulation()
        return simulation.result()

    def test_BoxComposition(self):
        """
        Compares simple box with the one composed from two smaller boxes of the same material to get same size
        """
        # reference box
        length = 50.0
        width = 20.0
        height = 10.0
        particle = Particle(particle_material,
                            FormFactorBox(length, width, height))
        particle.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - height/2))

        reference_data = self.get_result(particle)
        #IntensityDataIOFactory.writeIntensityData(reference_data, "ref_BoxComposition.int")

        # composition
        box = Particle(particle_material,
                       FormFactorBox(com_length/2, com_width, com_height))
        composition = ParticleComposition()
        # composition = ParticleComposition(box, positions)
        composition.addParticle(box, kvector_t(0, 0, 0))
        composition.addParticle(box, kvector_t(com_length/2, 0, 0))
        composition.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - com_height/2))

        data = self.get_result(composition)

        diff = RelativeDifference(data, reference_data)
        print("test_BoxComposition:", diff)
        self.assertLess(diff, 1e-10)

    def test_BoxCompositionRotateX(self):
        """
        Compares simple box with the one composed from two smaller boxes of the same material to get same size.
        Composed box is rotated around X, to get same orientation as the original box.
        """
        # reference box
        length = 50.0
        width = 10.0
        height = 20.0
        particle = Particle(particle_material,
                            FormFactorBox(length, width, height))
        particle.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - height/2))

        reference_data = self.get_result(particle)
        #IntensityDataIOFactory.writeIntensityData(reference_data, "ref_BoxCompositionRotateX.int")

        # composition
        box = Particle(particle_material,
                       FormFactorBox(com_length/2, com_width, com_height))
        composition = ParticleComposition()
        composition.addParticle(box, kvector_t(0, 0, 0))
        composition.addParticle(box, kvector_t(com_length/2, 0, 0))
        composition.setRotation(RotationX(90*deg))
        composition.setPosition(kvector_t(0, 0, -layer_thickness/2.))

        data = self.get_result(composition)

        diff = RelativeDifference(data, reference_data)
        print("test_BoxCompositionRotateX:", diff)
        self.assertLess(diff, 1e-10)

    def test_BoxCompositionRotateY(self):
        """
        Compares simple box with the one composed from two smaller boxes of the same material to get same size.
        Composed box is rotated around Y, to get same orientation as the original box.
        """
        # reference box
        length = 10.0
        width = 20.0
        height = 50.0
        particle = Particle(particle_material,
                            FormFactorBox(length, width, height))
        particle.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - height/2))

        reference_data = self.get_result(particle)
        #IntensityDataIOFactory.writeIntensityData(reference_data, "ref_BoxCompositionRotateY.int")

        # composition
        box = Particle(particle_material,
                       FormFactorBox(com_length/2, com_width, com_height))
        composition = ParticleComposition()
        composition.addParticle(box, kvector_t(0, 0, 0))
        composition.addParticle(box, kvector_t(com_length/2, 0, 0))
        composition.setRotation(RotationY(90*deg))
        composition.setPosition(
            kvector_t(0, 0, -layer_thickness/2. + com_length/4.))

        data = self.get_result(composition)

        diff = RelativeDifference(data, reference_data)
        print("test_BoxCompositionRotateY:", diff)
        self.assertLess(diff, 1e-10)

    def test_BoxCompositionRotateZ(self):
        """
        Compares simple box with the one composed from two smaller boxes of the same material to get same size.
        Composed box is rotated around Z, to get same orientation as the original box.
        """
        # reference box
        length = 20.0
        width = 50.0
        height = 10.0
        particle = Particle(particle_material,
                            FormFactorBox(length, width, height))
        particle.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - height/2))

        reference_data = self.get_result(particle)
        #IntensityDataIOFactory.writeIntensityData(reference_data, "ref_BoxCompositionRotateZ.int")

        # composition
        box = Particle(particle_material,
                       FormFactorBox(com_length/2, com_width, com_height))
        composition = ParticleComposition()
        composition.addParticle(box, kvector_t(0, 0, 0))
        composition.addParticle(box, kvector_t(com_length/2, 0, 0))
        composition.setRotation(RotationZ(90*deg))
        composition.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - com_height/2))

        data = self.get_result(composition)

        diff = RelativeDifference(data, reference_data)
        print("test_BoxCompositionRotateZ:", diff)
        self.assertLess(diff, 1e-10)

    def test_BoxCompositionRotateZandY(self):
        """
        Compares simple box with the one composed from two smaller boxes of the same material to get same size.
        Composed box is rotated around Y, to get same orientation as the original box.
        """
        # reference box
        length = 10.0
        width = 50.0
        height = 20.0
        particle = Particle(particle_material,
                            FormFactorBox(length, width, height))
        particle.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - height/2))

        reference_data = self.get_result(particle)
        #IntensityDataIOFactory.writeIntensityData(reference_data, "ref_BoxCompositionRotateZandY.int")

        # composition
        box = Particle(particle_material,
                       FormFactorBox(com_length/2, com_width, com_height))
        composition = ParticleComposition()
        composition.addParticle(box, kvector_t(0, 0, 0))
        composition.addParticle(box, kvector_t(com_length/2, 0, 0))
        composition.setRotation(RotationZ(90*deg))
        composition.rotate(RotationY(90*deg))
        composition.setPosition(kvector_t(0, 0, -layer_thickness/2.))

        data = self.get_result(composition)

        diff = RelativeDifference(data, reference_data)
        print("test_BoxCompositionRotateZandY:", diff)
        self.assertLess(diff, 1e-10)

    def test_BoxStackComposition(self):
        """
        2 different boxes are first rotated and then combined into the composition. Composition is then rotated.
        Composition is compared against reference box.
        """
        # reference box
        length = 10.0
        width = 20.0
        height = 50.0
        particle = Particle(particle_material,
                            FormFactorBox(length, width, height))
        particle.setPosition(
            kvector_t(0, 0, -layer_thickness/2.0 - height/2))
        reference_data = self.get_result(particle)
        #IntensityDataIOFactory.writeIntensityData(reference_data, "ref_BoxStackComposition.int")

        # composition
        composition = ParticleComposition()

        # box1 (20,50,5), rotatedZ
        box1_length = 20.0
        box1_width = 50.0
        box1_height = 5.0
        box1 = Particle(
            particle_material,
            FormFactorBox(box1_length, box1_width, box1_height))
        box1.setRotation(RotationZ(90*deg))

        # box2 (5,20,50), rotatedY
        box2_length = 5.0
        box2_width = 20.0
        box2_height = 50.0
        box2 = Particle(
            particle_material,
            FormFactorBox(box2_length, box2_width, box2_height))
        box2.setRotation(RotationY(90*deg))
        box2.setPosition(kvector_t(-box2_height/2, 0, box2_length/2))

        composition.addParticle(box1, kvector_t(0, 0, 0))
        composition.addParticle(box2, kvector_t(0, 0, box1_height))
        composition.setRotation(RotationY(90*deg))
        composition.setPosition(kvector_t(0, 0, -layer_thickness/2.))

        data = self.get_result(composition)

        diff = RelativeDifference(data, reference_data)
        print("test_BoxStackComposition:", diff)
        self.assertLess(diff, 1e-10)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""
This example demonstrates new experimental feature of release-1.8.0:
particles can now cross interfaces.

Given example is similar to simulation/ex01_BasicParticles/CylindersAndPrisms.py,
with the difference, that z-position of particles is adjusted to move them slightly
down from vacuum layer and to make them cross substrate/air interface.

Note:
1) Simulation kernel automatically detects particles crossing interface and
adjusts calculations accordingly.

2) Particle crossing interface requires more time to simulate.

3) Crossing of interface is possible only for limited number of geometries.
For example, X or Y rotated particles can not yet cross interfaces (exception
will be thrown when trying to simulate such geometries).
"""
import bornagain as ba
from bornagain import deg, nm, kvector_t


def get_sample():
    """
    Returns a sample with uncorrelated cylinders and prisms on a substrate.
    """

    # Define materials
    material_Particle = ba.HomogeneousMaterial("Particle", 0.0006, 2e-08)
    material_Substrate = ba.HomogeneousMaterial("Substrate", 6e-06, 2e-08)
    material_Vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)

    # Define form factors
    ff_1 = ba.FormFactorCylinder(5*nm, 5*nm)
    ff_2 = ba.FormFactorPrism3(10*nm, 5*nm)

    # Define particles
    particle_1 = ba.Particle(material_Particle, ff_1)
    particle_1_position = kvector_t(0, 0, -3*nm)
    particle_1.setPosition(particle_1_position)
    particle_2 = ba.Particle(material_Particle, ff_2)
    particle_2_position = kvector_t(0, 0, -3*nm)
    particle_2.setPosition(particle_2_position)

    # Define interference functions
    iff = ba.InterferenceFunctionNone()

    # Define particle layouts
    layout = ba.ParticleLayout()
    layout.addParticle(particle_1, 0.5)
    layout.addParticle(particle_2, 0.5)
    layout.setInterferenceFunction(iff)
    layout.setTotalParticleSurfaceDensity(0.01)

    # Define layers
    layer_1 = ba.Layer(material_Vacuum)
    layer_1.addLayout(layout)
    layer_2 = ba.Layer(material_Substrate)

    # Define sample
    sample = ba.MultiLayer()
    sample.addLayer(layer_1)
    sample.addLayer(layer_2)

    return sample


def get_simulation(sample):
    beam = ba.Beam(1, 0.1*nm, ba.Direction(0.2*deg, 0))
    detector = ba.SphericalDetector(100, 2*deg, 0, 1*deg)
    simulation = ba.GISASSimulation(beam, sample, detector)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



#!/usr/bin/env python3
"""
An example of taking into account beam angular and wavelength
divergence in reflectometry calculations with BornAgain.
"""
import bornagain as ba
from bornagain import angstrom, deg

# input parameters
wavelength = 1.54*angstrom
alpha_i_min = 0  # min incident angle, deg
alpha_i_max = 2*deg  # max incident angle, rad

# convolution parameters
d_wl = 0.01*wavelength  # spread width for wavelength
d_ang = 0.01*deg  # spread width for incident angle
n_sig = 3  # number of sigmas to convolve over
n_points = 25  # number of points to convolve over

# substrate (Si)
si_sld_real = 2.0704e-06  # \AA^{-2}
# layer parameters
n_repetitions = 10
# Ni
ni_sld_real = 9.4245e-06  # \AA^{-2}
d_ni = 70*angstrom  # ni layer thickness (nm)
# Ti
ti_sld_real = -1.9493e-06  # \AA^{-2}
d_ti = 30*angstrom  # ti layer thickness (nm)


def get_sample():
    # defining materials
    # this example implies beam divergence in the wavelength,
    # thus MaterialBySLD must be used to provide correct result
    m_vacuum = ba.MaterialBySLD("Vacuum", 0, 0)
    m_ni = ba.MaterialBySLD("Ni", ni_sld_real, 0)
    m_ti = ba.MaterialBySLD("Ti", ti_sld_real, 0)
    m_substrate = ba.MaterialBySLD("SiSubstrate", si_sld_real, 0)

    vacuum_layer = ba.Layer(m_vacuum)
    ni_layer = ba.Layer(m_ni, d_ni)
    ti_layer = ba.Layer(m_ti, d_ti)
    substrate_layer = ba.Layer(m_substrate)
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    for i in range(n_repetitions):
        multi_layer.addLayer(ti_layer)
        multi_layer.addLayer(ni_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation(sample, scan_size=500):
    """
    Returns a specular simulation with beam and detector defined.
    """

    alpha_distr = ba.RangedDistributionGaussian(n_points, n_sig)
    wavelength_distr = ba.RangedDistributionGaussian(n_points, n_sig)

    scan = ba.AngularSpecScan(wavelength, scan_size, alpha_i_min,
                              alpha_i_max)
    scan.setAbsoluteAngularResolution(alpha_distr, d_ang)
    scan.setAbsoluteWavelengthResolution(wavelength_distr, d_wl)

    simulation = ba.SpecularSimulation()
    simulation.setScan(scan)
    simulation.setSample(sample)

    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



#!/usr/bin/env python3
"""
Example of simulating a reflectometry experiment
with a rough sample using BornAgain.

"""
import bornagain as ba
from bornagain import deg, angstrom, nm


def get_sample():
    """
    Defines sample and returns it
    """

    # creating materials
    m_ambient = ba.MaterialBySLD("Ambient", 0, 0)
    m_ti = ba.MaterialBySLD("Ti", -1.9493e-06, 0)
    m_ni = ba.MaterialBySLD("Ni", 9.4245e-06, 0)
    m_substrate = ba.MaterialBySLD("SiSubstrate", 2.0704e-06, 0)

    # creating layers
    ambient_layer = ba.Layer(m_ambient)
    ti_layer = ba.Layer(m_ti, 30*angstrom)
    ni_layer = ba.Layer(m_ni, 70*angstrom)
    substrate_layer = ba.Layer(m_substrate)

    # defining roughness
    roughness = ba.LayerRoughness()
    roughness.setSigma(1*nm)

    # creating multilayer
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(ambient_layer)
    for i in range(10):
        multi_layer.addLayerWithTopRoughness(ti_layer, roughness)
        multi_layer.addLayerWithTopRoughness(ni_layer, roughness)
    multi_layer.addLayerWithTopRoughness(substrate_layer, roughness)

    return multi_layer


def get_simulation(sample, scan_size=500):
    """
    Defines and returns a specular simulation.
    """
    simulation = ba.SpecularSimulation()
    scan = ba.AngularSpecScan(1.54*angstrom, scan_size, 0, 2*deg)
    simulation.setScan(scan)
    simulation.setSample(sample)
    return simulation


if __name__ == '__main__':
    import ba_plot
    sample = get_sample()
    simulation = get_simulation(sample)
    ba_plot.run_and_plot(simulation)



#!/usr/bin/env python3
"""
Example of simulating a rough sample with a
tanh and Nevot-Croce roughness model using BornAgain.

"""

from matplotlib import pyplot as plt
import bornagain as ba
from bornagain import deg, angstrom, nm


def get_sample(roughness_model):
    """
    Defines sample and returns it
    """

    # creating materials
    m_ambient = ba.MaterialBySLD("Ambient", 0, 0)
    m_ti = ba.MaterialBySLD("Ti", -1.9493e-06, 0)
    m_ni = ba.MaterialBySLD("Ni", 9.4245e-06, 0)
    m_substrate = ba.MaterialBySLD("SiSubstrate", 2.0704e-06, 0)

    # creating layers
    ambient_layer = ba.Layer(m_ambient)
    ti_layer = ba.Layer(m_ti, 30*angstrom)
    ni_layer = ba.Layer(m_ni, 70*angstrom)
    substrate_layer = ba.Layer(m_substrate)

    # defining roughness
    roughness = ba.LayerRoughness()
    roughness.setSigma(1*nm)

    # creating multilayer
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(ambient_layer)
    for i in range(10):
        multi_layer.addLayerWithTopRoughness(ti_layer, roughness)
        multi_layer.addLayerWithTopRoughness(ni_layer, roughness)
    multi_layer.addLayerWithTopRoughness(substrate_layer, roughness)

    multi_layer.setRoughnessModel(roughness_model)

    return multi_layer


def get_simulation(sample, scan_size=500):
    """
    Defines and returns a specular simulation.
    """
    simulation = ba.SpecularSimulation()
    scan = ba.AngularSpecScan(1.54*angstrom, scan_size, 0, 2*deg)
    simulation.setScan(scan)
    simulation.setSample(sample)
    return simulation


def run_simulation(roughness_model=ba.RoughnessModel.TANH):
    """
    Runs simulation and returns its result.
    """
    sample = get_sample(roughness_model)
    simulation = get_simulation(sample)
    simulation.runSimulation()
    return simulation.result()


def plot(result_tanh, result_nevot_croce):

    plt.semilogy(result_nevot_croce.axis(),
                 result_nevot_croce.array(),
                 label="Névot-Croce")
    plt.semilogy(result_tanh.axis(), result_tanh.array(), label="Tanh")

    plt.xlabel(r'$\alpha_i \; (deg)$', fontsize=12)
    plt.ylabel(r'Intensity', fontsize=12)

    plt.legend()
    plt.show()


if __name__ == '__main__':
    result_tanh = run_simulation(roughness_model=ba.RoughnessModel.TANH)
    result_nevot_croce = run_simulation(
        roughness_model=ba.RoughnessModel.NEVOT_CROCE)
    plot(result_tanh, result_nevot_croce)



#!/usr/bin/env python3
"""
Fitting example: fit along slices
"""

from matplotlib import pyplot as plt
import bornagain as ba
import ba_plot
from bornagain import deg, angstrom, nm

phi_slice_value = 0.0  # position of vertical slice in deg
alpha_slice_value = 0.2  # position of horizontal slice in deg


def get_sample(params):
    """
    Returns a sample with uncorrelated cylinders on a substrate.
    """
    radius = params["radius"]
    height = params["height"]

    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_substrate = ba.HomogeneousMaterial("Substrate", 6e-6, 2e-8)
    m_particle = ba.HomogeneousMaterial("Particle", 6e-4, 2e-8)

    cylinder_ff = ba.FormFactorCylinder(radius, height)
    cylinder = ba.Particle(m_particle, cylinder_ff)

    layout = ba.ParticleLayout()
    layout.addParticle(cylinder)

    vacuum_layer = ba.Layer(m_vacuum)
    vacuum_layer.addLayout(layout)

    substrate_layer = ba.Layer(m_substrate, 0)
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayer(substrate_layer)
    return multi_layer


def get_simulation(params, add_masks=True):
    """
    Create and return GISAXS simulation with beam and detector defined
    """
    simulation = ba.GISASSimulation()
    simulation.setDetectorParameters(100, -1*deg, 1*deg, 100, 0, 2*deg)
    simulation.setBeamParameters(1*angstrom, 0.2*deg, 0)
    simulation.beam().setIntensity(1e+08)
    simulation.setSample(get_sample(params))
    if add_masks:
        """
        At this point we mask all the detector and then unmask two areas
        corresponding to the vertical and horizontal lines. This will make
        simulation/fitting to be performed along slices only.
        """
        simulation.maskAll()
        simulation.addMask(ba.HorizontalLine(alpha_slice_value*deg), False)
        simulation.addMask(ba.VerticalLine(phi_slice_value*deg), False)
    return simulation


def create_real_data():
    """
    Generating "real" data by adding noise to the simulated data.
    """
    # initial values which we will have to find later during the fit
    params = {'radius': 5*nm, 'height': 10*nm}

    # retrieving simulated data in the form of numpy array
    simulation = get_simulation(params, add_masks=False)
    simulation.setBackground(ba.PoissonNoiseBackground())
    simulation.runSimulation()

    return simulation.result().array()


class PlotObserver:
    """
    Draws fit progress every nth iteration. Here we plot slices along real
    and simulated images to see fit progress.
    """
    def __init__(self):
        self.fig = plt.figure(figsize=(10.25, 7.69))
        self.fig.canvas.draw()

    def __call__(self, fit_objective):
        self.update(fit_objective)

    @staticmethod
    def plot_real_data(data):
        """
        Plot experimental data as colormap with horizontal/vertical lines
        representing slices on top.
        """
        plt.subplots_adjust(wspace=0.2, hspace=0.2)
        ba_plot.plot_histogram(data, title="Experimental data")
        # line representing vertical slice
        plt.plot([phi_slice_value, phi_slice_value],
                 [data.getYmin(), data.getYmax()],
                 color='gray',
                 linestyle='-',
                 linewidth=1)
        # line representing horizontal slice
        plt.plot([data.getXmin(), data.getXmax()],
                 [alpha_slice_value, alpha_slice_value],
                 color='gray',
                 linestyle='-',
                 linewidth=1)

    @staticmethod
    def plot_slices(slices, title):
        """
        Plots vertical and horizontal projections.
        """
        plt.subplots_adjust(wspace=0.2, hspace=0.3)
        for label, slice in slices:
            plt.semilogy(slice.binCenters(),
                         slice.binValues(),
                         label=label)
            plt.xlim(slice.getXmin(), slice.getXmax())
            plt.ylim(1, slice.getMaximum()*10)
        plt.legend(loc='upper right')
        plt.title(title)

    @staticmethod
    def display_fit_parameters(fit_objective):
        """
        Displays fit parameters, chi and iteration number.
        """
        plt.title('Parameters')
        plt.axis('off')

        iteration_info = fit_objective.iterationInfo()

        plt.text(
            0.01, 0.85, "Iterations  " +
            '{:d}'.format(iteration_info.iterationCount()))
        plt.text(0.01, 0.75,
                 "Chi2       " + '{:8.4f}'.format(iteration_info.chi2()))
        for index, params in enumerate(iteration_info.parameters()):
            plt.text(
                0.01, 0.55 - index*0.1,
                '{:30.30s}: {:6.3f}'.format(params.name(), params.value))

        plt.tight_layout()
        plt.draw()
        plt.pause(0.01)

    def update(self, fit_objective):
        """
        Callback to access fit_objective on every n'th iteration.
        """
        self.fig.clf()

        real_data = fit_objective.experimentalData().histogram2d()
        simul_data = fit_objective.simulationResult().histogram2d()

        # plot real data
        plt.subplot(2, 2, 1)
        self.plot_real_data(real_data)

        # horizontal slices
        slices = [("real", real_data.projectionX(alpha_slice_value)),
                  ("simul", simul_data.projectionX(alpha_slice_value))]
        title = ("Horizontal slice at alpha =" +
                 '{:3.1f}'.format(alpha_slice_value))
        plt.subplot(2, 2, 2)
        self.plot_slices(slices, title)

        # vertical slices
        slices = [("real", real_data.projectionY(phi_slice_value)),
                  ("simul", simul_data.projectionY(phi_slice_value))]
        title = "Vertical slice at phi =" + '{:3.1f}'.format(
            phi_slice_value)
        plt.subplot(2, 2, 3)
        self.plot_slices(slices, title)

        # display fit parameters
        plt.subplot(2, 2, 4)
        self.display_fit_parameters(fit_objective)


def run_fitting():
    """
    main function to run fitting
    """

    real_data = create_real_data()

    fit_objective = ba.FitObjective()
    fit_objective.addSimulationAndData(get_simulation, real_data, 1)
    fit_objective.initPrint(10)

    # creating custom observer which will draw fit progress
    plotter = PlotObserver()
    fit_objective.initPlot(10, plotter)

    params = ba.Parameters()
    params.add("radius", 6.*nm, min=4, max=8)
    params.add("height", 9.*nm, min=8, max=12)

    minimizer = ba.Minimizer()
    result = minimizer.minimize(fit_objective.evaluate, params)
    fit_objective.finalize(result)


if __name__ == '__main__':
    run_fitting()
    plt.show()



#!/usr/bin/env python3
"""
Real life example:
Fitting data from an X-ray reflectometer

The sample is composed of a thin
silver nano-particle layer on a silicon
substrate. The substrate is covered with
SiO2 layer. The nano-particle layer has negligible density
and does not considerably affect
the observed reflectometry picture.

The following parameters of the experiment
are fitted:
1. beam intensity
2. footprint correction factor
3. beam angular divergence
4. Material concentration in the SiO2 layer
5. Thickness of SiO2 layer
6. Sample roughness

Fitting is done in two steps:
First the whole range of experimental data is fitted,
then the data related to the instrument is fixed and
on the second step only the right-hand part of
experimental data (i.e. the part of the reflectometry curve
associated with bigger incident angles)
is concerned for fitting. At the second
stage only the sample parameters are fitted,
since only they affect the shape of the reflectometry
curve at bigger incident angles.
"""
from matplotlib import pyplot as plt
import numpy as np
from os import path
import bornagain as ba
import ba_plot
from scipy.optimize import differential_evolution


def get_real_data(filename="mg6a_Merged.txt.gz"):
    """
    Loads real data files and merges them once.
    Returns a Nx3 array (N - the number of experimental data entries)
    with first column being coordinates,
    second one being values,
    and the third one being weights to restore intensity values from experiment
    """
    if not hasattr(get_real_data, "data"):
        filepath = path.join(path.dirname(path.realpath(__file__)),
                             filename)
        real_data = np.loadtxt(filepath, usecols=(0, 1, 3), skiprows=1)

        # translating axis values from double incident angle (degs)
        # to incident angle (radians)
        real_data[:, 0] *= np.pi/360
        get_real_data.data = real_data
    return get_real_data.data.copy()


def get_real_data_axis(start, end):
    """
    Get axis coordinates of the experimental data
    :param start: first bin to extract
    :param end: last bin to extract
    :return: 1D array with axis coordinates
    """
    return get_real_data()[start:end, 0]


def get_real_data_values(start, end):
    """
    Get experimental data values as a 1D array
    :param start: first bin to extract
    :param end: last bin to extract
    :return: 1D array with experimental data values
    """
    return get_real_data()[start:end, 1]


def get_weights(start, end):
    """
    Get weights to restore genuine intensity of experimental instrument
    :param start: first bin to extract
    :param end: last bin to extract
    :return: 1D array with weights to restore beam intensity
    """
    return get_real_data()[start:end, 2]


def create_simulation(arg_dict, bin_start, bin_end):
    """
    Creates and returns specular simulation
    """
    wavelength = 1.54*ba.angstrom
    alpha_distr = ba.RangedDistributionGaussian(30, 3)
    footprint = ba.FootprintGauss(arg_dict["footprint_factor"])

    scan = ba.AngularSpecScan(wavelength,
                              get_real_data_axis(bin_start, bin_end))
    scan.setAbsoluteAngularResolution(alpha_distr, arg_dict["divergence"])
    scan.setFootprintFactor(footprint)

    simulation = ba.SpecularSimulation()
    simulation.setScan(scan)
    simulation.beam().setIntensity(arg_dict["intensity"])
    return simulation


def buildSample(arg_dict):
    """
    Creates sample and returns it
    """
    # defining materials
    m_vacuum = ba.HomogeneousMaterial("Vacuum", 0, 0)
    m_si_o2 = ba.HomogeneousMaterial(
        "SiO2", 8.57040868e-06*arg_dict["concentration"],
        1.11016654e-07*arg_dict["concentration"])
    m_si = ba.HomogeneousMaterial("Si", 7.57211137e-06, 1.72728178e-07)

    # roughness
    r_si = ba.LayerRoughness(arg_dict["roughness"], 0, 0)

    # layers
    vacuum_layer = ba.Layer(m_vacuum)
    oxide_layer = ba.Layer(m_si_o2, arg_dict["thickness"])
    substrate_layer = ba.Layer(m_si)

    # assembling multilayer
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(vacuum_layer)
    multi_layer.addLayerWithTopRoughness(oxide_layer, r_si)
    multi_layer.addLayerWithTopRoughness(substrate_layer, r_si)

    return multi_layer


def run_simulation(arg_dict, bin_start=0, bin_end=-1):
    """
    Runs simulation and returns its result
    """

    simulation = create_simulation(arg_dict, bin_start, bin_end)
    simulation.setSample(buildSample(arg_dict))

    simulation.runSimulation()
    return simulation.result()


def chi_2(real_data, sim_data, weights):
    """
    Computes chi_2 metrics and returns its value
    """
    sim_data_upsc = np.multiply(weights, sim_data)
    sim_data_upsc[sim_data_upsc==0] = 1e-30
    real_data_upsc = np.multiply(weights, real_data)
    diff = real_data_upsc - sim_data_upsc
    return np.sum(np.divide(np.multiply(diff, diff), sim_data_upsc))


def create_par_dict(*arg):
    """
    Creates a dictionary with parameter names and values
    and returns it
    """
    return {
        'intensity': arg[0],
        'footprint_factor': arg[1],
        'divergence': arg[2],
        'concentration': arg[3],
        'thickness': arg[4],
        'roughness': arg[5]
    }


def objective_primary(args):
    """
    Objective function for preliminary stage of optimization
    """

    bin_start = 15  # first bin in the experimental data to calculate
    bin_end = -1  # last bin in the experimental data to calculate
    arg_dict = create_par_dict(*args)

    sim_result = run_simulation(arg_dict, bin_start, bin_end)
    sim_data = sim_result.array()
    return chi_2(get_real_data_values(bin_start, bin_end), sim_data,
                 get_weights(bin_start, bin_end))


def objective_fine(args, intensity, footprint_factor, divergence):
    """
    Objective function for tuning the right-hand side of experimental data
    """

    bin_start = 404  # first bin in the experimental data to calculate
    bin_end = -1  # last bin in the experimental data to calculate
    arg_dict = create_par_dict(intensity, footprint_factor, divergence,
                               *args)

    sim_result = run_simulation(arg_dict, bin_start, bin_end)
    sim_data = sim_result.array()
    return chi_2(get_real_data_values(bin_start, bin_end), sim_data,
                 get_weights(bin_start, bin_end))


def run_fitting():
    """
    Runs fitting and returns its result
    """
    # running preliminary optimization on the total range of experimental data.
    bounds = [
        (1e6, 1e8),  # beam intensity
        (0, 0.1),  # beam-to-sample width ratio
        (0, 0.08*ba.deg),  # beam_divergence
        (0, 1),  # oxide_concentration
        (0, 2*ba.nm),  # oxide_thickness
        (0, 2*ba.nm)
    ]  # roughness

    print("Start preliminary fitting of experimental data:\n")

    preliminary_result = differential_evolution(objective_primary,
                                                bounds,
                                                maxiter=20,
                                                popsize=60,
                                                mutation=(0.5, 1.5),
                                                disp=True,
                                                tol=1e-5)

    bounds = [
        (0, 1),  # oxide_concentration
        (0, 2*ba.nm),  # oxide_thickness
        (0, 2*ba.nm)
    ]  # roughness

    fixed_args = (
        preliminary_result.x[0],  # beam intensity
        preliminary_result.x[1],  # beam-to-sample width ratio
        preliminary_result.x[2]  # beam divergence
    )

    print(
        "\nStart fitting big incident angle part of experimental data:\n")

    fine_tuning_result = differential_evolution(objective_fine,
                                                bounds,
                                                fixed_args,
                                                maxiter=20,
                                                popsize=40,
                                                mutation=(0.5, 1.5),
                                                disp=True,
                                                tol=1e-5)

    result = create_par_dict(*fixed_args, *fine_tuning_result.x)
    print("\nFitting result:")
    print(result, "\n")

    return result


def plot_result(sim_result, ref_result, bin_start=0, bin_end=-1):
    """
    Plots the graphs of obtained simulation data
    """
    sim_data = sim_result.array()
    ref_data = ref_result.array()

    plt.semilogy(
        get_real_data_axis(bin_start, bin_end)*180/np.pi,
        get_real_data_values(bin_start, bin_end), sim_result.axis(),
        sim_data, ref_result.axis(), ref_data)

    xlabel = ba_plot.get_axes_labels(sim_result, ba.Axes.DEFAULT)[0]
    ylabel = "Intensity"
    plt.xlabel(xlabel, fontsize=16)
    plt.ylabel(ylabel, fontsize=16)
    plt.legend(['Experimental data', 'Simulation', 'Reference'],
               loc='upper right',
               fontsize=16)

    plt.show()


if __name__ == '__main__':
    fit_data = run_fitting()
    ref_data = create_par_dict(
        3.78271438e+06,  # beam intensity
        9.58009763e-04,  # beam-to-sample width ratio
        2.30471294e-04,  # beam angular divergence
        0.58721753,  # oxide concentration
        1.25559347,  # oxide thickness
        0.19281863)  # roughness
    plot_result(run_simulation(fit_data), run_simulation(ref_data))
